
from pathlib import Path

import click
import pydantic


__all__ = ('DEFAULT_CONFIG_DIR', 'Config', 'UserConfig')


DEFAULT_CONFIG_DIR = Path(click.get_app_dir('instawow'))


class Config(pydantic.BaseSettings):

    config_dir: Path
    addon_dir: Path

    @pydantic.validator('addon_dir')
    def expand_and_validate_addon_dir(cls, value: Path) -> Path:
        value = value.expanduser().resolve()
        if not value.is_dir():
            raise ValueError(f'{value} is not a directory')
        return value

    def write(self) -> None:
        "Create the profile."
        self.config_dir.mkdir(exist_ok=True)

    class Config:
        env_prefix = 'INSTAWOW_'


class UserConfig(Config):

    config_dir: Path = DEFAULT_CONFIG_DIR

    @classmethod
    def read(cls):  # -> UserConfig
        "Attempt to read ``addon_dir`` from disk."
        return cls(addon_dir=(cls(addon_dir='').config_dir/'addon_dir.txt')
                             .read_text(encoding='utf-8'))

    def write(self) -> None:
        "Create the profile and write ``addon_dir`` on disk."
        super().write()
        (self.config_dir/'addon_dir.txt').write_text(str(self.addon_dir),
                                                     encoding='utf-8')




from collections import namedtuple
from datetime import datetime
from pathlib import Path
import re
import typing as T

from . import __version__


__all__ = ('TocReader', 'slugify', 'is_outdated')


class TocReader:
    """Extracts key–value pairs from TOC files."""

    Entry = namedtuple('_TocEntry', 'key value')

    def __init__(self, path: Path) -> None:
        entries = (e.lstrip('# ').partition(': ')[::2]
                   for e in path.read_text(encoding='utf-8-sig').splitlines()
                   if e.startswith('## '))
        self.entries = dict(entries)

    def __getitem__(self, key: T.Union[str, T.Tuple[str]]) -> Entry:
        if isinstance(key, tuple):
            try:
                return next(filter(lambda i: i.value,
                                   (self.__getitem__(k) for k in key)))
            except StopIteration:
                key = key[0]
        return self.Entry(key, self.entries.get(key))


def slugify(text: str, *,
            _re_lc=re.compile(r'[^0-9a-z ]')) -> str:
    "Convert an add-on name into a lower-alphanumeric slug."
    return '-'.join(_re_lc.sub(' ', text.casefold()).split())


def is_outdated(manager) -> bool:
    """Check against PyPI to see if `instawow` is outdated.

    The response is cached for 24 hours.
    """
    cache_file = manager.config.config_dir/'.pypi_version'
    if cache_file.exists() and \
            (datetime.now() -
             datetime.fromtimestamp(cache_file.stat().st_mtime)).days < 1:
        version = cache_file.read_text(encoding='utf-8')
    else:
        from aiohttp.client import ClientError

        async def get_metadata():
            async with (await manager.client_factory()) as session, \
                       session.get('https://pypi.org/pypi/instawow/json') as response:
                return await response.json()

        try:
            version = manager.loop.run_until_complete(get_metadata())['info']['version']
        except ClientError:
            version = __version__
        else:
            cache_file.write_text(version, encoding='utf-8')
    return __version__ != version




from functools import partial, reduce
from itertools import count
from textwrap import fill
import typing as T
import webbrowser

import click
import logbook
from sqlalchemy import inspect, text
from texttable import Texttable

from . import __version__
from .config import UserConfig
from . import exceptions as E
from .manager import CliManager
from .models import Pkg, PkgFolder
from .utils import TocReader, is_outdated


__all__ = ('cli',)


_success = click.style('✓', fg='green')
_failure = click.style('✗', fg='red')
_warning = click.style('!', fg='blue')

MESSAGES = {
    E.PkgInstalled:
        f'{_success} {{0}}: installed {{1.new_pkg.version}}'
        f' from {{1.new_pkg.options.strategy}}'.format,
    E.PkgUpdated:
        f'{_success} {{0}}: updated {{1.old_pkg.version}} to {{1.new_pkg.version}}'
        f' from {{1.new_pkg.options.strategy}}'.format,
    E.PkgRemoved:
        f'{_success} {{}}: removed'.format,
    E.PkgAlreadyInstalled:
        f'{_failure} {{}}: already installed'.format,
    E.PkgConflictsWithInstalled:
        lambda a, r: f'{_failure} {a}: conflicts with installed '
                     f'add-on {_compose_addon_defn(r.conflicting_pkg)}',
    E.PkgConflictsWithPreexisting:
        f'{_failure} {{}}: conflicts with an add-on not installed by instawow\n'
         '  pass `-o` to `install` if you do actually wish to overwrite this add-on'
        .format,
    E.PkgNonexistent:
        f'{_failure} {{}}: no such project id or slug'.format,
    E.PkgTemporarilyUnavailable:
        f'{_failure} {{0}}: temporarily unavailable\n'
        f'  this usually means a new version of {{0}} is under review'.format,
    E.PkgNotInstalled:
        f'{_failure} {{}}: not installed'.format,
    E.PkgOriginInvalid:
        f'{_failure} {{}}: invalid origin'.format,
    E.PkgUpToDate:   # ignore
        None,
    E.InternalError:
        f'{_warning} {{}}: encountered {{.error.__class__.__name__}}'.format,}

_CONTEXT_SETTINGS = {'help_option_names': ['-h', '--help']}
_SEP = ':'


class _Parts(T.NamedTuple):

    origin: str
    id_or_slug: str


def _format_message(addon, result):
    return MESSAGES[result if isinstance(result, type)
                    else result.__class__](addon, result)


def _tabulate(rows: T.List[T.Tuple[str, ...]], *,
              head: T.Tuple[str, ...]=(), show_index: bool=True) -> str:
    table = Texttable(max_width=0).set_deco(Texttable.VLINES |
                                            Texttable.BORDER |
                                            Texttable.HEADER)
    if show_index:
        table.set_cols_align(f'r{"l" * len(rows[0])}')
        head = ('', *head)
        rows = [(i, *v) for i, v in enumerate(rows, start=1)]
    return table.add_rows([head, *rows]).draw()


def _compose_addon_defn(val: T.Union[Pkg, _Parts, tuple]) -> str:
    try:
        origin, slug = val.origin, val.slug
    except AttributeError:
        origin, slug = val
    return _SEP.join((origin, slug))


def _decompose_addon_defn(ctx, param, value, *,
                          raise_for_invalid_defn=True):
    if isinstance(value, tuple):
        return tuple(_decompose_addon_defn(ctx, param, v)
                     for v in sorted(set(value), key=value.index))

    for resolver in ctx.obj.resolvers.values():
        parts = resolver.decompose_url(value)
        if parts:
            parts = _Parts(*parts)
            break
    else:
        if _SEP not in value:
            if raise_for_invalid_defn:
                raise click.BadParameter(value)
            parts = _Parts('*', value)
        else:
            parts = _Parts(*value.partition(_SEP)[::2])
    return _compose_addon_defn(parts), parts


class _OrigCmdOrderGroup(click.Group):

    def list_commands(self, ctx):
        return self.commands    # The default is ``sorted(self.commands)``


def _init():
    addon_dir = click.prompt('Enter the path to your add-on folder')
    while True:
        try:
            UserConfig(addon_dir=addon_dir).write()
        except ValueError:
            addon_dir = click.prompt(f'{addon_dir} not found\n'
                                      'Enter the path to your add-on folder')
        else:
            break


@click.group(cls=_OrigCmdOrderGroup, context_settings=_CONTEXT_SETTINGS)
@click.version_option(__version__)
@click.option('--hide-progress', '-n', is_flag=True, default=False, hidden=True,
              help='Hide the progress bar.')
@click.pass_context
def main(ctx, hide_progress):
    """Add-on manager for World of Warcraft."""
    if not ctx.obj:
        for attempt in count():
            try:
                config = UserConfig.read()
            except (FileNotFoundError, ValueError):
                _init()
            else:
                break
        ctx.obj = manager = CliManager(config=config,
                                       show_progress=not hide_progress)
        if attempt:
            # Migrate add-on paths after redefining ``addon_dir``
            for folder in manager.db.query(PkgFolder).all():
                folder.path = manager.config.addon_dir/folder.path.name
            manager.db.commit()

        logbook.RotatingFileHandler(manager.config.config_dir/'error.log',
                                    delay=True)\
               .push_application()

        if is_outdated(manager):
            click.echo(f'{_warning} instawow is out of date:\n'
                       f'  run `python3 -m pip install -U instawow` to upgrade')

cli = main


@main.command()
@click.argument('addons', nargs=-1, required=True, callback=_decompose_addon_defn)
@click.option('--strategy', '-s',
              type=click.Choice(['canonical', 'latest']),
              default='canonical',
              help="Whether to install the latest published version "
                   "('canonical') or the very latest upload ('latest').")
@click.option('--overwrite', '-o',
              is_flag=True, default=False,
              help='Overwrite existing add-ons.')
@click.pass_obj
def install(manager, addons, overwrite, strategy):
    """Install add-ons."""
    for addon, result in zip((d for d, _ in addons),
                             manager.install_many((*p, strategy, overwrite)
                                                  for _, p in addons)):
        click.echo(_format_message(addon, result))


@main.command()
@click.argument('addons', nargs=-1, callback=_decompose_addon_defn)
@click.option('--strategy', '-s',
              type=click.Choice(['canonical', 'latest']), default=None,
              help="Whether to update to the latest published version "
                   "('canonical') or the very latest upload ('latest').")
@click.pass_obj
def update(manager, addons, strategy):
    """Update installed add-ons."""
    if not addons:
        addons = [(_compose_addon_defn(p), (p.origin, p.slug))
                  for p in manager.db.query(Pkg).all()]
    if strategy:
        for pkg in (manager.get(*p) for _, p in addons):
            pkg.options.strategy = strategy
        manager.db.commit()

    for addon, result in zip((d for d, _ in addons),
                             manager.update_many(p for _, p in addons)):
        if not isinstance(result, (E.PkgUpToDate,
                                   E.PkgTemporarilyUnavailable)):
            click.echo(_format_message(addon, result))


@main.command()
@click.argument('addons', nargs=-1, required=True, callback=_decompose_addon_defn)
@click.pass_obj
def remove(manager, addons):
    """Uninstall add-ons."""
    for addon, parts in addons:
        try:
            result = manager.remove(*parts)
        except (E.ManagerError, E.InternalError) as error:
            result = error
        click.echo(_format_message(addon, result))


@main.group('list')
def list_():
    """List add-ons."""


@list_.command('installed')
@click.option('--column', '-c', 'columns',
              multiple=True,
              help='A field to show in a column.  Nested fields are '
                   'dot-delimited.  Repeatable.')
@click.option('--columns', '-C', 'print_columns',
              is_flag=True, default=False,
              help='Print a list of all possible column values.')
@click.option('--toc-entry', '-t', 'toc_entries',
              multiple=True,
              help='An entry to extract from the TOC.  Repeatable.')
@click.option('--sort-by', '-s', 'sort_key',
              default='name',
              help='A key to sort the table by.  '
                   'You can chain multiple keys by separating them with a comma '
                   'just as you would in SQL, '
                   'e.g. `--sort-by="origin, date_published DESC"`.')
@click.pass_obj
def list_installed(manager,
                   columns, print_columns, toc_entries,
                   sort_key):
    """List installed add-ons."""
    def format_columns(pkg):
        for column in columns:
            try:
                value = reduce(getattr, [pkg] + column.split('.'))
            except AttributeError:
                raise click.BadParameter(column,
                                         param_hint=['--column', '-c'])
            if column == 'folders':
                value = '\n'.join(f.path.name for f in value)
            elif column == 'options':
                value = {'strategy': value.strategy}
            elif column == 'description':
                value = fill(value, width=50, max_lines=3)
            yield value

    def format_toc_entries(pkg):
        toc_readers = [TocReader(f.path/f'{f.path.name}.toc')
                       for f in pkg.folders]
        for toc_entry in toc_entries:
            value = [fill(r[toc_entry].value or '', width=50)
                     for r in toc_readers]
            value = sorted(set(value), key=value.index)
            value = '\n'.join(value)
            yield value

    if print_columns:
        click.echo(_tabulate([(c,) for c in (*inspect(Pkg).columns.keys(),
                                             *inspect(Pkg).relationships.keys())],
                             head=('field',), show_index=False))
    else:
        click.echo(_tabulate([(_compose_addon_defn(p),
                               *format_columns(p),
                               *format_toc_entries(p))
                              for p in manager.db.query(Pkg)
                                                 .order_by(text(sort_key))
                                                 .all()],
                             head=('add-on',
                                   *columns,
                                   *(f'[{e}]' for e in toc_entries))))


@list_.command('outdated')
@click.option('--flip', '-f', 'should_flip',
              is_flag=True, help='Check against the opposing strategy.')
@click.pass_obj
def list_outdated(manager, should_flip):
    """List outdated add-ons."""
    def flip(strategy):
        if should_flip:
            strategy = 'latest' if strategy == 'canonical' else 'canonical'
        return strategy

    outdated = manager.db.query(Pkg).order_by(Pkg.name).all()
    outdated = zip(outdated, manager.resolve_many((p.origin, p.id,
                                                   flip(p.options.strategy))
                                                  for p in outdated))
    outdated = [(i, n) for i, n in outdated if i.file_id != n.file_id]
    if outdated:
        click.echo(_tabulate([(_compose_addon_defn(n),
                               i.version, n.version, n.options.strategy)
                              for i, n in outdated],
                             head=('add-on', 'installed', 'new', 'strategy')))


@list_.command('preexisting')
@click.pass_obj
def list_preexisting(manager):
    """List add-ons not installed by instawow."""
    folders = {f.name
               for f in manager.config.addon_dir.iterdir() if f.is_dir()} - \
              {f.path.name for f in manager.db.query(PkgFolder).all()}
    folders = ((n, manager.config.addon_dir/n/f'{n}.toc') for n in folders)
    folders = {(n, TocReader(t)) for n, t in folders if t.exists()}
    if folders:
        click.echo(_tabulate([(n,
                               t['X-Curse-Project-ID'].value or '',
                               t['X-WoWI-ID'].value or '',
                               t['X-Curse-Packaged-Version', 'X-Packaged-Version',
                                 'Version'].value or '')
                              for n, t in sorted(folders)],
                             head=('folder', 'Curse ID', 'WoWI ID', 'version')))


@main.command()
@click.argument('addon', callback=partial(_decompose_addon_defn,
                                          raise_for_invalid_defn=False))
@click.option('--toc-entry', '-t', 'toc_entries',
              multiple=True,
              help='An entry to extract from the TOC.  Repeatable.')
@click.pass_obj
def info(manager, addon, toc_entries):
    """Show the details of an installed add-on."""
    pkg = addon[1]
    pkg = (manager.get(*pkg) or manager.db.query(Pkg)
                                          .filter(Pkg.slug.contains(pkg.id_or_slug))
                                          .order_by(Pkg.name)
                                          .first())
    if pkg:
        rows = {'name': pkg.name,
                'source': pkg.origin,
                'id': pkg.id,
                'slug': pkg.slug,
                'description': fill(pkg.description, max_lines=5),
                'homepage': pkg.url,
                'version': pkg.version,
                'release date': pkg.date_published,
                'folders': '\n'.join([str(pkg.folders[0].path.parent)] +
                                     [' ├─ ' + f.path.name for f in pkg.folders[:-1]] +
                                     [' └─ ' + pkg.folders[-1].path.name]),
                'strategy': pkg.options.strategy,}

        if toc_entries:
            for path, toc_reader in ((f.path,
                                      TocReader(f.path/f'{f.path.name}.toc'))
                                     for f in pkg.folders):
                rows.update({f'[{path.name} {k}]': fill(toc_reader[k].value or '')
                             for k in toc_entries})
        click.echo(_tabulate(rows.items(), show_index=False))
    else:
        click.echo(_format_message(addon[0], E.PkgNotInstalled))


@main.command()
@click.argument('addon', callback=partial(_decompose_addon_defn,
                                          raise_for_invalid_defn=False))
@click.pass_obj
def hearth(manager, addon):
    """Open the add-on's homepage in your browser."""
    pkg = addon[1]
    pkg = (manager.get(*pkg) or manager.db.query(Pkg)
                                          .filter(Pkg.slug.contains(pkg.id_or_slug))
                                          .order_by(Pkg.name)
                                          .first())
    if pkg:
        webbrowser.open(pkg.url)
    else:
        click.echo(_format_message(addon[0], E.PkgNotInstalled))


@main.command()
@click.argument('addon', callback=partial(_decompose_addon_defn,
                                          raise_for_invalid_defn=False))
@click.pass_obj
def reveal(manager, addon):
    """Open the add-on folder in your file manager."""
    pkg = addon[1]
    pkg = (manager.get(*pkg) or manager.db.query(Pkg)
                                          .filter(Pkg.slug.contains(pkg.id_or_slug))
                                          .order_by(Pkg.name)
                                          .first())
    if pkg:
        webbrowser.open(pkg.folders[0].path.as_uri())
    else:
        click.echo(_format_message(addon[0], E.PkgNotInstalled))


@main.group()
def extras():
    """Additional functionality."""


@extras.group()
def bitbar():
    """Mini update GUI implemented as a BitBar plug-in.

    BitBar <https://getbitbar.com/> is a menu-bar app host for macOS.
    """


@bitbar.command('_generate', hidden=True)
@click.argument('caller')
@click.argument('version')
@click.pass_obj
def bitbar_generate(manager, caller, version):
    from base64 import b64encode
    from pathlib import Path
    from subprocess import run, PIPE

    outdated = manager.db.query(Pkg).order_by(Pkg.name).all()
    outdated = zip(outdated,
                   manager.resolve_many((p.origin, p.id, p.options.strategy)
                                        for p in outdated))
    outdated = [(p, r) for p, r in outdated
                if p.file_id != getattr(r, 'file_id', p.file_id)]

    icon = Path(__file__).parent/'assets'/f'''\
NSStatusItem-icon__\
{"has-updates" if outdated else "clear"}\
{"@2x" if b"Retina" in run(("system_profiler", "SPDisplaysDataType"),
                           stdout=PIPE).stdout else ""}\
.png'''
    icon = b64encode(icon.read_bytes()).decode()

    click.echo(f'''| templateImage="{icon}"
---
instawow ({__version__}:{version})''')
    if outdated:
        if len(outdated) > 1:
            click.echo(f'''\
Update all | bash={caller} param1=update terminal=false refresh=true''')
        for p, r in outdated:
            click.echo(f'''\
Update “{r.name}” ({p.version} ➞ {r.version}) | \
  refresh=true \
  bash={caller} param1=update param2={_compose_addon_defn(r)} terminal=false
View “{r.name}” on {manager.resolvers[r.origin].name} | \
  alternate=true \
  bash={caller} param1=hearth param2={_compose_addon_defn(r)} terminal=false''')


@bitbar.command('install')
def bitbar_install():
    """Install the instawow BitBar plug-in."""
    from pathlib import Path
    import tempfile
    import sys

    with tempfile.TemporaryDirectory() as name:
        path = Path(name, 'instawow.1h.py')
        path.write_text(f'''\
#!/usr/bin/env LC_ALL=en_US.UTF-8 {sys.executable}

__version__ = {__version__!r}

import sys

from instawow.cli import main

main(['-n', *(sys.argv[1:] or ['extras', 'bitbar', '_generate', sys.argv[0], __version__])])
''')
        webbrowser.open(f'bitbar://openPlugin?src={path.as_uri()}')
        click.pause('Press any key to exit after installing the plug-in')




import asyncio
from contextlib import ExitStack
import contextvars
from functools import partial
import io
from pathlib import Path
import typing as T
import zipfile

from aiohttp import ClientSession, TCPConnector, TraceConfig
import logbook
from send2trash import send2trash
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from tqdm import tqdm

from .config import Config
from . import exceptions as E
from .models import ModelBase, Pkg, PkgFolder
from .resolvers import CurseResolver, WowiResolver, TukuiResolver


__all__ = ('Manager', 'CliManager')


logger = logbook.Logger(__name__)


_UA_STRING = 'instawow (https://github.com/layday/instawow)'

_client = contextvars.ContextVar('_client')


def _init_loop():
    try:
        import uvloop
    except ImportError:
        return asyncio.get_event_loop()
    else:
        return uvloop.new_event_loop()


def _init_db_session(*, config):
    engine = create_engine(f'sqlite:///{config.config_dir/"db.sqlite"}')
    ModelBase.metadata.create_all(engine)
    return sessionmaker(bind=engine)()


async def _init_web_client(*, loop, **kwargs):
    return ClientSession(loop=loop,
                         connector=TCPConnector(loop=loop, limit_per_host=10),
                         headers={'User-Agent': _UA_STRING},
                         **kwargs)


class PkgArchive:

    __slots__ = ('archive', 'root_folders')

    def __init__(self, payload: bytes) -> None:
        self.archive = zipfile.ZipFile(io.BytesIO(payload))

        folders = sorted({Path(p).parts[0] for p in self.archive.namelist()})
        self.root_folders = [Path(p) for p in folders]

    def extract(self, parent_folder: Path, *,
                overwrite: bool=False) -> None:
        "Extract the archive contents under ``parent_folder``."
        if not overwrite:
            conflicts = {f.name for f in self.root_folders} & \
                        {f.name for f in parent_folder.iterdir()}
            if conflicts:
                raise E.PkgConflictsWithPreexisting(conflicts)
        self.archive.extractall(parent_folder)


class MemberDict(dict):

    def __missing__(self, key):
        raise E.PkgOriginInvalid(origin=key)


class Manager:

    def __init__(self, *,
                 config: Config, loop: asyncio.AbstractEventLoop=None,
                 client_factory: T.Callable=None) -> None:
        self.config = config
        self.loop = loop or _init_loop()
        self.client_factory = partial(client_factory or _init_web_client,
                                      loop=self.loop)
        self.client = _client
        self.db = _init_db_session(config=self.config)
        self.resolvers = MemberDict((r.origin, r(manager=self))
                                    for r in (CurseResolver, WowiResolver, TukuiResolver))

    def get(self, origin: str, id_or_slug: str) -> Pkg:
        "Retrieve a ``Pkg`` from the database."
        return self.db.query(Pkg)\
                      .filter(Pkg.origin == origin,
                              (Pkg.id == id_or_slug) | (Pkg.slug == id_or_slug))\
                      .first()

    async def resolve(self, origin: str, id_or_slug: str, strategy: str) -> Pkg:
        "Resolve an ID or slug into a ``Pkg``."
        return await self.resolvers[origin].resolve(id_or_slug,
                                                    strategy=strategy)

    async def to_install(self, origin: str, id_or_slug: str, strategy: str,
                         overwrite: bool) -> T.Callable[[], E.PkgInstalled]:
        "Retrieve a package to install."
        def install():
            archive = PkgArchive(payload)
            pkg.folders = [PkgFolder(path=self.config.addon_dir/f)
                           for f in archive.root_folders]

            conflicts = self.db.query(PkgFolder)\
                               .filter(PkgFolder.path.in_
                                        ([f.path for f in pkg.folders]))\
                               .first()
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts.pkg)

            if overwrite:
                for path in (f.path for f in pkg.folders if f.path.exists()):
                    send2trash(str(path))
            archive.extract(parent_folder=self.config.addon_dir,
                            overwrite=overwrite)
            self.db.add(pkg)
            self.db.commit()
            return E.PkgInstalled(pkg)

        if self.get(origin, id_or_slug):
            raise E.PkgAlreadyInstalled
        pkg = await self.resolve(origin, id_or_slug, strategy)
        async with self.client.get()\
                              .get(pkg.download_url) as response:
            payload = await response.read()
        return install

    async def to_update(self, origin: str,
                        id_or_slug: str) -> T.Callable[[], E.PkgUpdated]:
        "Retrieve a package to update."
        def update():
            archive = PkgArchive(payload)
            new_pkg.folders = [PkgFolder(path=self.config.addon_dir/f)
                               for f in archive.root_folders]

            conflicts = self.db.query(PkgFolder)\
                               .filter(PkgFolder.path.in_
                                        ([f.path for f in new_pkg.folders]),
                                       PkgFolder.pkg_origin != new_pkg.origin,
                                       PkgFolder.pkg_id != new_pkg.id)\
                               .first()
            if conflicts:
                raise E.PkgConflictsWithInstalled(conflicts.pkg)

            with ExitStack() as stack:
                stack.callback(self.db.commit)
                for folder in old_pkg.folders:
                    send2trash(str(folder.path))
                self.db.delete(old_pkg)
                archive.extract(parent_folder=self.config.addon_dir)
                self.db.add(new_pkg)
            return E.PkgUpdated(old_pkg, new_pkg)

        old_pkg = self.get(origin, id_or_slug)
        if not old_pkg:
            raise E.PkgNotInstalled
        new_pkg = await self.resolve(origin, id_or_slug, old_pkg.options.strategy)
        if old_pkg.file_id == new_pkg.file_id:
            raise E.PkgUpToDate

        async with self.client.get()\
                              .get(new_pkg.download_url) as response:
            payload = await response.read()
        return update

    def remove(self, origin: str, id_or_slug: str) -> E.PkgRemoved:
        "Remove a package."
        pkg = self.get(origin, id_or_slug)
        if not pkg:
            raise E.PkgNotInstalled

        for folder in pkg.folders:
            send2trash(str(folder.path))
        self.db.delete(pkg)
        self.db.commit()
        return E.PkgRemoved(pkg)


class Bar(tqdm):

    def __init__(self, *args, **kwargs) -> None:
        kwargs['position'], self.__reset_position = kwargs['position']
        super().__init__(*args, **{'leave': False, 'ascii': True, **kwargs})

    def close(self) -> None:
        super().close()
        self.__reset_position()


async def _init_cli_web_client(*, loop, manager):
    async def do_on_request_end(session, context, params):
        if params.response.content_type in {
                'application/zip',
                # Curse at it again
                'application/x-amz-json-1.0'}:
            filename = params.response.headers.get('Content-Disposition', '')
            filename = filename[(filename.find('"') + 1):filename.rfind('"')] or \
                       params.response.url.name
            bar = Bar(total=params.response.content_length,
                      desc=f'  Downloading {filename}',
                      miniters=1, unit='B', unit_scale=True,
                      position=manager.bar_position)

            async def ticker(bar=bar, params=params):
                while True:
                    if params.response.content._eof:
                        bar.close()
                        break
                    bar.update(params.response.content._cursor - bar.n)
                    await asyncio.sleep(bar.mininterval)
            loop.create_task(ticker())

    trace_conf = TraceConfig()
    trace_conf.on_request_end.append(do_on_request_end)
    trace_conf.freeze()
    return await _init_web_client(loop=loop, trace_configs=[trace_conf])


class CliManager(Manager):

    def __init__(self, *,
                 config: Config, loop: asyncio.AbstractEventLoop=None,
                 show_progress: bool=True) -> None:
        super().__init__(config=config, loop=loop,
                         client_factory=(partial(_init_cli_web_client, manager=self)
                                         if show_progress else None))
        self.show_progress = show_progress
        self._bar_position = []

    @property
    def bar_position(self) -> T.Tuple[int, T.Callable]:
        "Get the first available bar slot."
        try:
            b = self._bar_position.index(False)
            self._bar_position[b] = True
        except ValueError:
            b = len(self._bar_position)
            self._bar_position.append(True)
        return (b, lambda b=b: self._bar_position.__setitem__(b, False))

    def run(self, awaitable: T.Awaitable) -> T.Any:
        "Run ``awaitable`` inside an explicit context."
        async def runner():
            async with (await self.client_factory()) as client:
                _client.set(client)
                return await awaitable

        return contextvars.copy_context().run(partial(self.loop.run_until_complete,
                                                      runner()))

    async def gather(self, it: T.Iterable, **kwargs) -> list:
        async def intercept_exc(position, awaitable):
            try:
                result = await awaitable
            except E.ManagerError as error:
                result = error
            except Exception as error:
                logger.exception()
                result = E.InternalError(error=error)
            return position, result

        futures = [intercept_exc(*i) for i in enumerate(it)]
        results = [None] * len(futures)
        with Bar(total=len(futures), disable=not self.show_progress,
                 position=self.bar_position, **kwargs) as bar:
            for result in asyncio.as_completed(futures, loop=self.loop):
                results.__setitem__(*await result)
                bar.update(1)
            # Wait for ``ticker``s to complete so all bars get to wipe
            # their pretty selves off the screen
            while len(asyncio.all_tasks(self.loop)) > 1:
                await asyncio.sleep(bar.mininterval)
        return results

    @staticmethod
    def _intercept_exc(callable_):
        try:
            result = callable_()
        except E.ManagerError as error:
            result = error
        except Exception as error:
            logger.exception()
            result = E.InternalError(error=error)
        return result

    def resolve_many(self, values: T.Iterable) -> list:
        results = self.run(self.gather((self.resolve(*a) for a in values),
                                       desc='Resolving'))
        return results

    def install_many(self, values: T.Iterable) -> list:
        results = self.run(self.gather((self.to_install(*a) for a in values),
                                       desc='Fetching'))
        return [self._intercept_exc(i) for i in results]

    def update_many(self,  values: T.Iterable) -> list:
        results = self.run(self.gather((self.to_update(*a) for a in values),
                                       desc='Checking'))
        return [self._intercept_exc(i) for i in results]
